#!/usr/bin/env python3

"""
 Copyright 2024 - The Minton Group at Purdue University
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

"""
Generates and runs a set of Swiftest input files from initial conditions for the Spherical Harmonics features with the WHM integrator. 
"""

import swiftest
import numpy as np

seed = 123
rng = np.random.default_rng(seed=seed)


# Central Body Parameters (just an oblate sphere to test) 
cb_mass = 6.1e18 # kg
cb_a = 160 # km
cb_b = 160 # km
cb_c = 90 # km
cb_volume = 4.0 / 3 * np.pi * cb_a*cb_b*cb_c**3 # km^3
cb_density = cb_mass / cb_volume 
cb_T_rotation = 7.004 / 24.0 # converting from hours to julian days (TU)
cb_rot = [[0, 0, 360.0 / cb_T_rotation]] # degrees/d

# Add 1 user-defined test particle.
ntp = 1

name_tp     = ["TestParticle_01"]
a_tp        = 400
e_tp        = 0.05
inc_tp      = 10
capom_tp    = 0.0
omega_tp    = 0.0
capm_tp     = 0.0


# Add 1 user-defined massive particle
npl = 1
density_pl = cb_density

name_pl     = ["MassiveBody_01"]
a_pl        = 300.0
e_pl        = 0.03
inc_pl      = 0.001
capom_pl    = 90.0
omega_pl    = 90.0
capm_pl     = 90.0
R_pl        = 1.0
M_pl        = 4.0 / 3 * np.pi * R_pl**3 * density_pl
Ip_pl       = np.full((npl,3),0.4,)
rot_pl      = np.zeros((npl,3))
mtiny       = 0.1 * np.max(M_pl)


# Extract the spherical harmonics coefficients (c_lm) from axes measurements
#
# The user can pass an optional reference radius at which the coefficients are calculated. If not provided, SHTOOLS 
# calculates the reference radius. If lref_radius = True, the function returns the reference radius used. 
# We recommend setting passing and setting a reference radius. Coefficients are geodesy (4-pi) normalised.

c_lm, cb_radius = swiftest.clm_from_ellipsoid(mass = cb_mass, density = cb_density, a = cb_a, b = cb_b, c = cb_c, lmax = 6, lref_radius = True)

# extracting only the J2 terms
tmp20 = c_lm[0, 2, 0] # c_20 = -J2
c_lm = np.zeros(np.shape(c_lm))
c_lm[0, 2, 0] = tmp20

J2 = -tmp20 * np.sqrt(5) # unnormalised J2 term
j2rp2 = J2 * cb_radius**2

# set up swiftest simulation with relevant units (here they are km, days, and kg)
sim_shgrav = swiftest.Simulation(simdir="shgrav",DU2M = 1e3, TU = 'd', MU = 'kg')

sim_shgrav.clean()
# Use the shgrav version where you input a set of spherical harmonics coefficients
sim_shgrav.add_body(name = 'OblateBody', mass = cb_mass, rot = cb_rot, radius = cb_radius, c_lm = c_lm)
sim_shgrav.add_body(name=name_tp, a=a_tp, e=e_tp, inc=inc_tp, capom=capom_tp, omega=omega_tp, capm=capm_tp)
sim_shgrav.add_body(name=name_pl, a=a_pl, e=e_pl, inc=inc_pl, capom=capom_pl, omega=omega_pl, capm=capm_pl, mass=M_pl, radius=R_pl,  Ip=Ip_pl, rot=rot_pl)
sim_shgrav.run(tstart=0.0, tstop=10.0, dt=0.01, tstep_out=10.0, dump_cadence=0, mtiny=mtiny, integrator='symba')

# Use the original "oblate" version where you pass J2 (and/or J4)
sim_obl = swiftest.Simulation(simdir="obl", DU2M = 1e3, TU='d', MU='kg')
sim_obl.clean()
sim_obl.add_body(name = 'OblateBody', mass = cb_mass, rot = cb_rot, radius = cb_radius, J2 = j2rp2)
sim_obl.add_body(name=name_tp, a=a_tp, e=e_tp, inc=inc_tp, capom=capom_tp, omega=omega_tp, capm=capm_tp)
sim_obl.add_body(name=name_pl, a=a_pl, e=e_pl, inc=inc_pl, capom=capom_pl, omega=omega_pl, capm=capm_pl, mass=M_pl, radius=R_pl,  Ip=Ip_pl, rot=rot_pl)
sim_obl.run(tstart=0.0, tstop=10.0, dt=0.01, tstep_out=10.0, dump_cadence=0, mtiny=mtiny, integrator='symba')

diff_vars = ['a','e','inc','capom','omega','capm','rh','vh']
ds_diff = sim_shgrav.data[diff_vars] - sim_obl.data[diff_vars]
ds_diff /= sim_obl.data[diff_vars]

print(ds_diff.isel(time=-1,name=-2))
print(ds_diff.isel(time=-1,name=-1))




"""
Copyright 2022 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
This file is part of Swiftest.
Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Swiftest. 
If not, see: https://www.gnu.org/licenses. 
"""

import numpy as np
import xarray as xr
from scipy.spatial.transform import Rotation as R

def magnitude(ds,x):
    """
    Computes the magnitude of a vector quantity from a Dataset.
    
    Parameters
    ----------
    ds : Xarray Dataset
        Dataset containing the vector quantity
    x : str
        Name of the vector quantity variable in the Dataset, which must have a "space" dimension (x,y,z coordinates)
    """
    dim = "space"
    ord = None
    return xr.apply_ufunc(
        np.linalg.norm, ds[x].where(~np.isnan(ds[x])), input_core_dims=[[dim]], kwargs={"ord": ord, "axis": -1}, dask="allowed"
    )
    
        
def wrap_angle(angle):
    """
    Converts angles to be between 0 and 360 degrees.
        
    Parameters
    ----------
    angle : float
        Angle to be converted
    Returns
    -------
    angle : float
        Converted angle
    """
    while np.any(angle >= 360.0 ):
        angle[angle >= 360.0] -= 360.0
    while np.any(angle < 0.0):
        angle[angle < 0.0] += 360.0
    return angle


def follow_swift(ds, ifol=None, nskp=None):
    """
    Emulates the Swift version of tool_follow.f
    It will generate a file called follow.out containing 10 columns (angles are all in degrees)::
    
        1 2 3  4    5     6    7    8    9   10
        t,a,e,inc,capom,omega,capm,peri,apo,obar
        
    Parameters
    ----------
    ds : Xarray Dataset 
        Dataset containing orbital elements
    ifol : int, optional
        Particle number to follow. The default is None, in which case the user is prompted to enter a particle number.
    nskp : int, optional
        Print frequency. The default is None, in which case the user is prompted to enter a print frequency. 

    Returns
    -------
    fol : Xarray Dataset
        Dataset containing only the followed body with angles converted to degrees

    """
    fol = None
    if ifol is None:
        intxt = input(' Input the particle number to follow \n')
        ifol = int(intxt)
    print(f"Following particle {ifol}")
    if ifol < 0: # Negative numbers are planets
        fol = ds.where(np.invert(np.isnan(ds['Gmass'])), drop=True)
        fol = fol.where(np.invert(np.isnan(fol['a'])), drop=True) # Remove times where this body doesn't exist (but this also gets rid of the central body)
        fol = fol.isel(id = -ifol - 2)  # Take 1 off for 0-indexed arrays in Python, and take 1 more off because the central body is gone
    elif ifol > 0: # Positive numbers are test particles
        fol = ds.where(np.isnan(ds['Gmass']), drop=True).drop_vars(['Gmass', 'radius'])
        fol = fol.where(np.invert(np.isnan(fol['a'])), drop=True)
        fol = fol.isel(id = ifol - 1)  # Take 1 off for 0-indexed arrays in Python

    if nskp is None:
        intxt = input('Input the print frequency\n')
        nskp = int(intxt)
        
    fol['obar'] = fol['capom'] + fol['omega']
    fol['obar'] = fol['obar'].fillna(0)
    fol['obar'] = wrap_angle(fol['obar'])
    fol['peri'] = fol['a'] * (1.0 - fol['e'])
    fol['apo']  = fol['a'] * (1.0 + fol['e'])

    
    tslice = slice(None, None, nskp)
    try:
        with open('follow.out', 'w') as f:
            print('# 1 2 3  4    5     6    7    8    9   10', file=f)
            print('# t,a,e,inc,capom,omega,capm,peri,apo,obar', file=f)
            for t in fol.isel(time=tslice).time:
                a = fol['a'].sel(time=t).values
                e = fol['e'].sel(time=t).values
                inc = fol['inc'].sel(time=t).values
                capom = fol['capom'].sel(time=t).values
                omega = fol['omega'].sel(time=t).values
                capm = fol['capm'].sel(time=t).values
                peri = fol['peri'].sel(time=t).values
                apo = fol['apo'].sel(time=t).values
                obar = fol['obar'].sel(time=t).values
                print(f"{t.values:15.7e} {a:22.16f} {e:22.16f} {inc:22.16f} {capom:22.16f} {omega:22.16f} {capm:22.16f} {peri:22.16f} {apo:22.16f} {obar:22.16f}", file=f)
                
    except IOError:
        print(f"Error writing to follow.out")
    
    return fol


def danby(M, ecc, accuracy=1e-14):
    """
    Danby's method to solve Kepler's equation. See [1]_ and [2]_ for details.

    Parameters
    ----------
    M : float
        the mean anomaly in radians
    ecc : float
        the eccentricity
    accuracy : float
        the relative accuracy to obtain a solution. Default is 1e-12.

    Returns
    ----------
    E : float
        the eccentric anomaly in radians

    References
    __________
    .. [1] Danby, J.M.A. 1988. Fundamentals of celestial mechanics. Richmond, Va., U.S.A., Willmann-Bell, 1988. 2nd ed.
    .. [2] Murray, C.D., Dermott, S.F., 1999. Solar system dynamics, New York, New York. ed, Cambridge University Press.

    """

    def kepler_root(E, ecc, M, deriv):
        """
        The Kepler equation root function. The returned value depends on the value of `deriv`, where::
        
            deriv = 0 : 
                E - e * np.sin(E) - M
            deriv = 1 : 
                1 - e * np.cos(E)
            deriv = 2 : 
                e * np.sin(E)
            deriv = 3 : 
                e * np.cos(E)

        The function will return 0 when E is correct for a given e and M

        Parameters
        ----------
        E : float
            the eccentric anomaly in radians
        ecc : float
            the eccentricity
        M : float
            the mean anomaly in radians
        deriv : int between 0 and 3
            The order of the derivative to compute

        Returns
        ----------
        float
            The value of the Kepler equation root function
        """

        if deriv == 0:
            return E - ecc * np.sin(E) - M
        elif deriv == 1:
            return 1.0 - ecc * np.cos(E)
        elif deriv == 2:
            return ecc * np.sin(E)
        elif deriv == 3:
            return ecc * np.cos(E)
        else:
            print(f"deriv={deriv} is undefined")
            return None

    def delta_ij(E, ecc, M, j):
        """
        Danby's intermediate delta functions. This function is recursive for j>1.

        Parameters
        ----------
        E : float
            the eccentric anomaly in radians
        ecc : float
            the eccentricity
        M : float
            the mean anomaly in radians
        j : int between 1 and 3
            The order of the delta function to compute

        Returns
        ----------
        float:
            delta_ij value used in Danby's iterative Kepler equation solver
        """
        if j == 1:
            return -kepler_root(E, ecc, M, 0) / kepler_root(E, ecc, M, 1)
        elif j == 2:
            return -kepler_root(E, ecc, M, 0) / (kepler_root(E, ecc, M, 1)
                                                 - delta_ij(E, ecc, M, 1) * kepler_root(E, ecc, M, 2) / 2.0)
        elif j == 3:
            return -kepler_root(E, ecc, M, 0) / (kepler_root(E, ecc, M, 1)
                                                 + delta_ij(E, ecc, M, 1) * kepler_root(E, ecc, M, 2) / 2.0
                                                 + delta_ij(E, ecc, M, 2) ** 2 * kepler_root(E, ecc, M,
                                                                                             3) / 6.0)
        else:
            print(f"j = {j} is not a valid input to the delta_ij function")

    # If this is a circular orbit, just return the mean anomaly as the eccentric anomaly
    if ecc < np.finfo(np.float64).tiny:  # Prevent floating point exception for 0 eccentricity orbits
        return M

    # Initial guess
    k = 0.85
    E = M + np.sign(np.sin(M)) * k * ecc
    MAXLOOPS = 50  # Maximum number of times to iterate before we give up
    for i in range(MAXLOOPS):
        Enew = E + delta_ij(E, ecc, M, 3)
        if np.abs((Enew - E) / E) < accuracy:
            return Enew
        E = Enew

    raise RuntimeError("The danby function did not converge on a solution.")


def el2xv_one(mu, a, ecc, inc, Omega, omega, M):
    """
    Compute osculating orbital elements from relative Cartesian position and velocity.
    All angular measures are returned in radians
    If inclination < TINY, longitude of the ascending node is arbitrarily set to 0
    If eccentricity < sqrt(TINY), argument of pericenter is arbitrarily set to 0

    ALGORITHM -  See Fitzpatrick "Principles of Cel. Mech."

    Adapted from Martin Duncan's el2xv.f

    Parameters
    ----------
    mu : float
        Central body gravitational constant
    a : float
        semimajor axis
    ecc : float
        eccentricity
    inc : float
        inclination (degrees)
    Omega : float
        longitude of ascending node (degrees)
    omega : float
        argument of periapsis (degrees)
    M : flaot
        Mean anomaly (degrees)

    Returns
    ----------
    rvec : (3) float vector
        Cartesian position vector
    vvec : (3) float vector
        Cartesian velocity vector
    """

    if ecc < 0.0:
        print("Error in el2xv! Eccentricity cannot be negative. Setting it to 0")
        e = 0.0
        iorbit_type = "ellipse"
    else:
        e = ecc
        em1 = e - 1.
        if np.abs(em1) < 2 * np.finfo(float).eps:
            iorbit_type = "parabola"
        elif e > 1.0:
            iorbit_type = "hyperbola"
        else:
            iorbit_type = "ellipse"

    def scget(angle):
        """
        Efficiently compute the sine and cosine of an input angle
        Input angle must be in radians

        Adapted from David E. Kaufmann's Swifter routine: orbel_scget.f90
        Adapted from Hal Levison's Swift routine orbel_scget.f

        Parameters
        ----------
        angle : float
            input angle in radians

        Returns
        -------
        sx : float
            sin of angle
        cx : float
            cos of angle

        """
        TWOPI = 2 * np.pi
        nper = int(angle / TWOPI)
        x = angle - nper * TWOPI
        if x < 0.0:
            x += TWOPI

        sx = np.sin(x)
        cx = np.sqrt(1.0 - sx ** 2)
        if x > np.pi / 2.0 and x < 3 * np.pi / 2.0:
            cx = -cx
        return sx, cx

    sip, cip = scget(np.deg2rad(omega))
    so, co = scget(np.deg2rad(Omega))
    si, ci = scget(np.deg2rad(inc))

    d11 = cip * co - sip * so * ci
    d12 = cip * so + sip * co * ci
    d13 = sip * si
    d21 = -sip * co - cip * so * ci
    d22 = -sip * so + cip * co * ci
    d23 = cip * si

    # Get the other quantities depending on orbit type
    if iorbit_type == "ellipse":
        E = danby(np.deg2rad(M),e)
        scap, ccap = scget(E)
        sqe = np.sqrt(1. - e**2)
        sqgma = np.sqrt(mu * a)
        xfac1 = a * (ccap - e)
        xfac2 = a * sqe * scap
        ri = 1. / (a * (1. - e * ccap))
        vfac1 = -ri * sqgma * scap
        vfac2 = ri * sqgma * sqe * ccap
    else:
        print(f"Orbit type: {iorbit_type} not yet implemented")
        xfac1 = 0.0
        xfac2 = 0.0
        vfac1 = 0.0
        vfac2 = 0.0

    rvec = np.array([d11 * xfac1 + d21 * xfac2,
                     d12 * xfac1 + d22 * xfac2,
                     d13 * xfac1 + d23 * xfac2])
    vvec = np.array([d11 * vfac1 + d21 * vfac2,
                     d12 * vfac1 + d22 * vfac2,
                     d13 * vfac1 + d23 * vfac2])

    return rvec, vvec


def el2xv_vec(mu, a, ecc, inc, Omega, omega, M):
    """
    Vectorized call to el2xv_one
    
    Parameters
    ----------
    mu : float
        Central body gravitational constant
    a : (n) float array
        semimajor axis
    ecc : (n) float array
        eccentricity
    inc : (n) float array
        inclination (degrees)
    Omega : (n) float array
        longitude of ascending node (degrees)
    omega : (n) float array
        argument of periapsis (degrees)
    M : (n) float array
        Mean anomaly (degrees)

    Returns
    ----------
    rvec : (n,3) float array
        Cartesian position vector
    vvec : (n,3) float array
        Cartesian velocity vector
    """
    vecfunc = np.vectorize(el2xv_one, signature='(),(),(),(),(),(),()->(3),(3)')
    return vecfunc(mu, a, ecc, inc, Omega, omega, M)


def xv2el_one(mu,rvec,vvec):
    """
    Converts from cartesian position and velocity values to orbital elements

    Parameters
    ----------
    mu : float
        Central body gravitational constant
    rvec : (3) float array
        Cartesian position vector
    vvec : (3) float array
        Cartesian velocity vector

    Returns
    ----------
    a : float
        semimajor axis
    ecc : float
        eccentricity
    inc : float
        inclination (degrees)
    Omega : float
        longitude of ascending node (degrees)
    omega : float
        argument of periapsis (degrees)
    M : float
        mean anomaly (degrees)
    varpi : float
        longitude of periapsis (degrees)
    f : float
        true anomaly (degrees)
    lam : float
        mean longitude (degrees)
    """

    rmag = np.linalg.norm(rvec)
    vmag2 = np.vdot(vvec,vvec)
    h = np.cross(rvec,vvec)
    hmag = np.linalg.norm(h)

    rdot = np.sign(np.vdot(rvec,vvec)) * np.sqrt(vmag2 - (hmag / rmag)**2)

    a = 1.0/(2.0 / rmag - vmag2/mu)
    ecc = np.sqrt(1 - hmag**2 / (mu * a))
    inc = np.arccos(h[2]/hmag)

    goodinc = np.abs(inc) > np.finfo(np.float64).tiny
    sO = np.where(goodinc,  np.sign(h[2]) * h[0] / (hmag * np.sin(inc)),0.0)
    cO = np.where(goodinc, -np.sign(h[2]) * h[1] / (hmag * np.sin(inc)),1.0)

    Omega = np.arctan2(sO, cO)

    sof = np.where(goodinc,rvec[2] / (rmag * np.sin(inc)), rvec[1]/rmag)
    cof = np.where(goodinc,(rvec[0] / rmag + np.sin(Omega) * sof * np.cos(inc)) / np.cos(Omega), rvec[0]/rmag)

    of = np.arctan2(sof,cof)

    goodecc = ecc > np.finfo(np.float64).tiny
    sf = np.where(goodecc,a * (1.0 - ecc**2)  * rdot / (hmag * ecc), 0.0)
    cf = np.where(goodecc,(1.0 / ecc) * (a * (1.0 - ecc**2) / rmag - 1.0), 1.0 )

    f = np.arctan2(sf, cf)

    omega = of - f

    varpi = Omega + omega

    # Compute eccentric anomaly & mean anomaly in order to get mean longitude
    E = np.where(ecc > np.finfo(np.float64).tiny, np.arccos(-(rmag - a) / (a * ecc)), 0)
    if np.sign(np.vdot(rvec, vvec)) < 0.0:
        E = 2 * np.pi - E

    M = E - ecc * np.sin(E)
    lam = M + varpi

    return a, ecc, np.rad2deg(inc), np.rad2deg(Omega), np.rad2deg(omega), np.rad2deg(M), np.rad2deg(varpi), np.rad2deg(f), np.rad2deg(lam)


def xv2el_vec(mu, rvec, vvec):
    """
    Vectorized call to xv2el_one.

    Parameters
    ----------
    mu : float
        Central body gravitational constant
    rvec : (n,3) float array
        Cartesian position vector
    vvec : (n,3) float array
        Cartesian velocity vector

    Returns
    ----------
    a : (n) float array
        semimajor axis
    ecc : (n) float array
        eccentricity
    inc : (n) float array
        inclination (degrees)
    Omega : (n) float array
        longitude of ascending node (degrees)
    omega : (n) float array
        argument of periapsis (degrees)
    M : (n) float array
        mean anomaly (degrees)
    varpi : (n) flaot array
        longitude of periapsis (degrees)
    f : (n) float array
        true anomaly (degrees)
    lam : (n) float array
        mean longitude (degrees)
    """

    vecfunc = np.vectorize(xv2el_one, signature='(),(3),(3)->(),(),(),(),(),(),(),(),()')
    return vecfunc(mu, rvec, vvec)


def rotate_to_vector(ds, new_pole, skip_vars=['space','Ip']):
    """
    Rotates the coordinate system such that the z-axis is aligned with an input pole.  The new pole is defined by the input vector. 
    This will change all variables in the Dataset that have the "space" dimension, except for those passed to the skip_vars parameter.
    
    Parameters
    ----------
    ds : Xarray Dataset
        Dataset containing the vector quantity
    new_pole : (3) float array
        New pole vector
    skip_vars : list of str, optional
        List of variable names to skip. The default is ['space','Ip'].
        
    Returns
    -------
    ds : Xarray Dataset
        Dataset with the new pole vector applied to all variables with the "space" dimension
    """
    
    if 'space' not in ds.dims:
        print("No space dimension in Dataset")
        return ds
    
    # Verify that the new pole is a 3-element array
    if len(new_pole) != 3:
        print("New pole must be a 3-element array")
        return ds
  
    # Normalize the new pole vector to ensure it is a unit vector
    pole_mag = np.linalg.norm(new_pole)
    unit_pole = new_pole / pole_mag
    
    # Define the original and target vectors
    target_vector = np.array([0, 0, 1])  # Rotate so that the z-axis is aligned with the new pole
    original_vector = unit_pole.reshape(1, 3)  
    
    # Use align_vectors to get the rotation that aligns the z-axis with Mars_rot
    rotation, _ = R.align_vectors(target_vector, original_vector)

    # Define a function to apply the rotation, which will be used with apply_ufunc
    def apply_rotation(vector, rotation):
        return rotation.apply(vector)
    
    # Function to apply rotation to a DataArray
    def rotate_dataarray(da, rotation):
        return xr.apply_ufunc(
            apply_rotation,
            da,
            kwargs={'rotation': rotation},
            input_core_dims=[['space']],
            output_core_dims=[['space']],
            vectorize=True,
            dask='parallelized',
            output_dtypes=[da.dtype]
        )

    # Loop through each variable in the dataset and apply the rotation if 'space' dimension is present
    for var in ds.variables:
        if 'space' in ds[var].dims and var not in skip_vars:
            ds[var] = rotate_dataarray(ds[var], rotation)
            
    return ds
        
        


#!/usr/bin/env python3

"""
 Copyright 2023 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

"""
Reads and processes a Swiftest output file and a Swifter output file.

Inputs
_______
param.swifter.in  : An ASCII file containing the parameters for the Swifter simulation.
param.swiftest.in : An ASCII file containing the parameters for the Swiftest simulation.
bin.dat           : A binary file containing the Swifter simulation output.
data.nc           : A NetCDF file containing the Swiftest simulation output.

Returns
-------
swifter_swiftest_comp.eps : Encapsulated PostScript containing the number of massive bodies and test particles for a Swifter and Swiftest run over time.
"""

import swiftest
import numpy as np
import matplotlib.pyplot as plt
import xarray as xr

# Pull in the Swifter Data
swifter_sim = swiftest.Simulation(param_file="param.swifter.in", codename="Swifter").ds

# Pull in the Swiftest Data
swiftest_sim = swiftest.Simulation(param_file="param.swiftest.in").ds

# Number of Bodies in Swifter System
npl_swifter = []
ntp_swifter = []
for i in range(len(swifter_sim['time'])):
    npl_swifter.append(len(swifter_sim['Gmass'][i].values[np.logical_not(np.isnan(swifter_sim['Gmass'][i].values))]))
    tp_only = swifter_sim.where(swifter_sim['id'] >= 159)
    ntp_swifter.append(len(tp_only['a'][i].values[np.logical_not(np.isnan(tp_only['a'][i].values))]))

# Calculate Slope of Decay
npl_swifter_slope = np.polyfit(swifter_sim['time'], npl_swifter, 1)[0]
ntp_swifter_slope = np.polyfit(swifter_sim['time'], ntp_swifter, 1)[0]
npl_swiftest_slope = np.polyfit(swiftest_sim['time'], swiftest_sim['npl'], 1)[0]
ntp_swiftest_slope = np.polyfit(swiftest_sim['time'], swiftest_sim['ntp'], 1)[0]

# Plot
fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8,4))
ax.plot(swifter_sim['time'], npl_swifter, color='crimson', linestyle='solid', linewidth=1, label=f'Swifter Massive Bodies\nslope={npl_swifter_slope}')
ax.plot(swiftest_sim['time'], swiftest_sim['npl'], color='navy', linestyle='solid', linewidth=1, label=f'Swiftest Massive Bodies\nslope={npl_swiftest_slope}')
ax.plot(swifter_sim['time'], ntp_swifter, color='crimson', linestyle='dotted', linewidth=1, label=f'Swifter Test Particles\nslope={ntp_swifter_slope}')
ax.plot(swiftest_sim['time'], swiftest_sim['ntp'], color='navy', linestyle='dotted', linewidth=1, label=f'Swiftest Test Particles\nslope={ntp_swiftest_slope}')
ax.set_title("Total Number of Bodies in the System")
ax.set_xlabel("Time (y)", fontsize=12)
ax.set_ylabel("n", fontsize=12)
ax.set_xlim(0,1e6)
ax.set_xticks([0, 2e5, 4e5, 6e5, 8e5, 1e6])
ax.set_xticklabels(['0','$2 * 10^5$','$4 * 10^5$','$6 * 10^5$','$8 * 10^5$', '$1 * 10^6$'], size=12)
ax.set_ylim(40,110)
ax.set_yticks([40, 50, 60, 70, 80, 90, 100, 110])
ax.set_yticklabels(['40', '50', '60', '70', '80', '90', '100', '110'], size=12)
ax.tick_params(direction="in", which='both')
plt.legend(fontsize=12, frameon=False, bbox_to_anchor=(1,1), loc="upper left")
plt.tight_layout()
plt.savefig(f'swifter_swiftest_comp.eps', dpi=300,facecolor='white', transparent=False)
plt.show()


#!/usr/bin/env python3

"""
 Copyright 2023 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

"""
Generates and runs a set of Swiftest input files and a set of Swifter input files from initial conditions with the 
SyMBA integrator. 

Input
------
None.

Output
------
param.swiftest.in      : An ASCII file containing the parameters for the Swiftest simulation.
param.swifter.in       : An ASCII file containing the parameters for the Swifter simulation.
cb.swiftest.in         : An ASCII file containing the central body data for the Swiftest simulation.
pl.swifter.in          : An ASCII file containing the massive body data for the Swifter simulation.
pl.swiftest.in         : An ASCII file containing the massive body data for the Swiftest simulation.
tp.swifter.in          : An ASCII file containing the test particle data for the Swifter simulation.
tp.swiftest.in         : An ASCII file containing the test particle data for the Swiftest simulation.

"""

import numpy as np
import swiftest
import swiftest.io as swio
import astropy.constants as const
import sys
import xarray as xr

# Both codes use the same tp input file
tpin = "tp.in"

swifter_input  = "param.swifter.in"
swifter_pl     = "pl.swifter.in"
swifter_tp     = "tp.swifter.in"
swifter_bin    = "bin.swifter.dat"
swifter_enc    = "enc.swifter.dat"

swiftest_input = "param.swiftest.in"
swiftest_pl    = "pl.swiftest.in"
swiftest_tp    = "tp.swiftest.in"
swiftest_cb    = "cb.swiftest.in"
swiftest_bin   = "bin.swiftest.nc"
swiftest_enc   = "enc.swiftest.dat"
swiftest_dis  = "discard.swiftest.dat"

sim = swiftest.Simulation()

sim.param['T0'] = 0.0
sim.param['DT'] = 1.0 
sim.param['TSTOP'] = 365.25e1
sim.param['ISTEP_OUT']  = 10
sim.param['ISTEP_DUMP'] = 10
sim.param['CHK_QMIN_COORD'] = "HELIO"
sim.param['CHK_QMIN'] = swiftest.RSun / swiftest.AU2M
sim.param['CHK_QMIN_RANGE'] = f"{swiftest.RSun / swiftest.AU2M} 1000.0"
sim.param['CHK_RMIN'] = swiftest.RSun / swiftest.AU2M
sim.param['CHK_RMAX'] = 1000.0
sim.param['CHK_EJECT'] = 1000.0
sim.param['OUT_STAT'] = "UNKNOWN"
sim.param['GR'] = 'NO'
sim.param['CHK_CLOSE'] = 'YES'
sim.param['RHILL_PRESENT'] = 'YES'
sim.param['GMTINY'] = 1.0e-12
sim.param['IN_FORM'] = 'XV'

sim.param['MU2KG'] = swiftest.MSun
sim.param['TU2S'] = swiftest.JD2S
sim.param['DU2M'] = swiftest.AU2M

bodyid = {
   "Sun": 0,
   "Mercury": 1,
   "Venus": 2,
   "Earth": 3,
   "Mars": 4,
   "Jupiter": 5,
   "Saturn": 6,
   "Uranus": 7,
   "Neptune": 8,
}

for name, id in bodyid.items():
   sim.add(name, idval=id)

ds = sim.ds
cb = ds.sel(id=0)
pl = ds.where(ds.id > 0, drop=True)
npl = pl.id.size

ntp = 16
dims = ['time', 'id', 'vec']
tp = []
t = np.array([0.0])
sim.param['OUT_FORM'] = "XV"
clab, plab, tlab = swio.make_swiftest_labels(sim.param)

# For each planet, we will initialize a pair of test particles. One on its way in, and one on its way out. We will also initialize two additional particles that don't encounter anything
tpidlist = np.arange(9,9+ntp)
tpnames = []
for i in tpidlist:
   tpnames.append(f"SmallBody{i+1:0.4g}")
tpxv1 = np.empty((6))
tpxv2 = np.empty((6))

p1 = []
p2 = []
p3 = []
p4 = []
p5 = []
p6 = []

for i in pl.id:
    pli = pl.sel(id=i)
    rstart = 2 * np.double(pli['radius'])  # Start the test particles at a multiple of the planet radius away
    vstart = 1.5 * np.sqrt(2 * np.double(pli['Gmass'])  / rstart)  # Start the test particle velocities at a multiple of the escape speed
    xvstart = np.array([rstart / np.sqrt(2.0), rstart / np.sqrt(2.0), 0.0, vstart, 0.0, 0.0])
    # The positions and velocities of each pair of test particles will be in reference to a planet
    plvec = np.array([np.double(pli['xhx']),
                      np.double(pli['xhy']),
                      np.double(pli['xhz']),
                      np.double(pli['vhx']),
                      np.double(pli['vhy']),
                      np.double(pli['vhz'])])
    tpxv1 = plvec + xvstart
    tpxv2 = plvec - xvstart
    p1.append(tpxv1[0])
    p1.append(tpxv2[0])
    p2.append(tpxv1[1])
    p2.append(tpxv2[1])
    p3.append(tpxv1[2])
    p3.append(tpxv2[2])
    p4.append(tpxv1[3])
    p4.append(tpxv2[3])
    p5.append(tpxv1[4])
    p5.append(tpxv2[4])
    p6.append(tpxv1[5])
    p6.append(tpxv2[5])


tvec = np.vstack([p1, p2, p3, p4, p5, p6])
tpframe = np.expand_dims(tvec.T, axis=0)
tpxr = xr.DataArray(tpframe, dims = dims, coords = {'time' : t, 'id' : tpidlist, 'vec' : tlab})
tpxr = tpxr.assign_coords(name=('id', tpnames))

tp = [tpxr]
tpda = xr.concat(tp,dim='time')
tpds = tpda.to_dataset(dim = 'vec')

sim.ds = xr.combine_by_coords([sim.ds, tpds])
swio.swiftest_xr2infile(sim.ds, sim.param)

sim.param['PL_IN'] = swiftest_pl
sim.param['TP_IN'] = swiftest_tp
sim.param['CB_IN'] = swiftest_cb
sim.param['BIN_OUT'] = swiftest_bin
sim.param['ENC_OUT'] = swiftest_enc
sim.param['DISCARD_OUT'] = swiftest_dis  
sim.save(swiftest_input)

sim.param['PL_IN'] = swifter_pl
sim.param['TP_IN'] = swifter_tp
sim.param['BIN_OUT'] = swifter_bin
sim.param['ENC_OUT'] = swifter_enc
sim.param['OUT_TYPE'] = "REAL8"
sim.save(swifter_input, codename="Swifter")



#!/usr/bin/env python3
"""
 Copyright 2023 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

"""
Generates and runs a set of Swiftest input files from initial conditions with the SyMBA integrator. All simulation 
outputs are stored in the subdirectory named after their collisional regime. 

**NOTE: You must have ffmpeg installed on your system before running this script. For instance, on MacOS:

```brew install ffmpeg```

on Ubuntu:

```sudo apt-get install ffmpeg```


Inputs
_______
User selects which collisional movie to generate

Output
------

Files are stored in directories for each collision type name requested
[COLLISION_TYPE].mp4    : A movie file named after the collisional regime depicting the collision.

"""

import swiftest
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from scipy.spatial.transform import Rotation as R

# ----------------------------------------------------------------------------------------------------------------------
# Define the names and initial conditions of the various fragmentation simulation types
# ----------------------------------------------------------------------------------------------------------------------
available_movie_styles = ["disruption_headon", "disruption_off_axis", "supercatastrophic_headon", "supercatastrophic_off_axis","hitandrun_disrupt", "hitandrun_pure", "merge", "merge_spinner"]
movie_title_list = ["Head-on Disruption", "Off-axis Disruption", "Head-on Supercatastrophic", "Off-axis Supercatastrophic", "Hit and Run w/ Runner Disruption", "Pure Hit and Run", "Merge", "Merge crossing the spin barrier"]
movie_titles = dict(zip(available_movie_styles, movie_title_list))
num_movie_frames = 1000


# ----------------------------------------------------------------------------------------------------------------------
# To increase the number of bodies generated in each collision type, decrease the value of the corresponding nfrag_reduction number 
# ----------------------------------------------------------------------------------------------------------------------
nfrag_reduction = {"disruption_headon" : 10.0,
         "disruption_off_axis"         : 10.0,
         "supercatastrophic_headon"    : 10.0,
         "supercatastrophic_off_axis"  : 10.0,
         "hitandrun_disrupt"           : 10.0,
         "hitandrun_pure"              : 1.0,
         "merge"                       : 1.0,
         "merge_spinner"               : 1.0,
         }



# These initial conditions were generated by trial and error
names = ["Target","Projectile"]
pos_vectors = {"disruption_headon"         : [np.array([1.0, -5.0e-05, 0.0]),
                                              np.array([1.0,  5.0e-05 ,0.0])],
              "disruption_off_axis"        : [np.array([1.0, -5.0e-05, 0.0]),
                                              np.array([1.0,  5.0e-05 ,0.0])], 
               "supercatastrophic_headon":   [np.array([1.0, -5.0e-05, 0.0]),
                                              np.array([1.0,  5.0e-05, 0.0])],
               "supercatastrophic_off_axis": [np.array([1.0, -5.0e-05, 0.0]),
                                              np.array([1.0,  5.0e-05, 0.0])],
               "hitandrun_disrupt"         : [np.array([1.0, -4.2e-05, 0.0]),
                                              np.array([1.0,  4.2e-05, 0.0])],
               "hitandrun_pure"            : [np.array([1.0, -4.2e-05, 0.0]),
                                              np.array([1.0,  4.2e-05, 0.0])],
               "merge"                      : [np.array([1.0, -5.0e-05, 0.0]),
                                              np.array([1.0,  5.0e-05 ,0.0])],
               "merge_spinner"               : [np.array([1.0, -5.0e-05, 0.0]),
                                              np.array([1.0,  5.0e-05 ,0.0])]                
               }

vel_vectors = {"disruption_headon"         : [np.array([ 0.00,  6.280005, 0.0]),
                                              np.array([ 0.00,  3.90,     0.0])],
               "disruption_off_axis"       : [np.array([ 0.00,  6.280005, 0.0]),
                                              np.array([ 0.05,  3.90,     0.0])],
               "supercatastrophic_headon":   [np.array([ 0.00,  6.28,     0.0]),
                                              np.array([ 0.00,  4.28,     0.0])],
               "supercatastrophic_off_axis": [np.array([ 0.00,  6.28,     0.0]),
                                              np.array([ 0.05,  4.28,     0.0])],
               "hitandrun_disrupt"         : [np.array([ 0.00,  6.28,     0.0]),
                                              np.array([-1.45, -6.28,     0.0])],
               "hitandrun_pure"            : [np.array([ 0.00,  6.28,     0.0]),
                                              np.array([-1.52, -6.28,     0.0])],
               "merge"                     : [np.array([ 0.04,  6.28,     0.0]),
                                              np.array([ 0.05,  6.18,     0.0])],
               "merge_spinner"             : [np.array([ 0.04,  6.28,     0.0]),
                                              np.array([ 0.05,  6.18,     0.0])] 
               }

rot_vectors = {"disruption_headon"         : [np.array([0.0, 0.0, 1.0e5]),
                                              np.array([0.0, 0.0, -5e5])],
               "disruption_off_axis":        [np.array([0.0, 0.0, 2.0e5]),
                                              np.array([0.0, 0.0, -1.0e5])],
               "supercatastrophic_headon":   [np.array([0.0, 0.0, 1.0e5]),
                                              np.array([0.0, 0.0, -5.0e5])],
               "supercatastrophic_off_axis": [np.array([0.0, 0.0, 1.0e5]),
                                              np.array([0.0, 0.0, -1.0e4])],
               "hitandrun_disrupt"         : [np.array([0.0, 0.0, 0.0]),
                                              np.array([0.0, 0.0, 1.0e5])],
               "hitandrun_pure"            : [np.array([0.0, 0.0, 0.0]),
                                              np.array([0.0, 0.0, 1.0e5])],
               "merge"                     : [np.array([0.0, 0.0, 0.0]),
                                              np.array([0.0, 0.0, 0.0])],
               "merge_spinner"             : [np.array([0.0, 0.0, -1.2e6]),
                                              np.array([0.0, 0.0, 0.0])],
               }

body_Gmass = {"disruption_headon"        : [1e-7, 1e-9],
             "disruption_off_axis"       : [1e-7, 1e-9],
             "supercatastrophic_headon"  : [1e-7, 1e-8],
             "supercatastrophic_off_axis": [1e-7, 1e-8],
             "hitandrun_disrupt"         : [1e-7, 7e-10],
             "hitandrun_pure"            : [1e-7, 7e-10],
             "merge"                     : [1e-7, 1e-8],
             "merge_spinner"             : [1e-7, 1e-8] 
               }

tstop = {"disruption_headon"         : 2.0e-3,
         "disruption_off_axis"       : 2.0e-3,
         "supercatastrophic_headon"  : 2.0e-3,
         "supercatastrophic_off_axis": 2.0e-3,
         "hitandrun_disrupt"         : 2.0e-4,
         "hitandrun_pure"            : 2.0e-4,
         "merge"                     : 5.0e-3,
         "merge_spinner"             : 5.0e-3,
         }

density = 3000 * swiftest.AU2M**3 / swiftest.MSun
GU = swiftest.GMSun * swiftest.YR2S**2 / swiftest.AU2M**3
body_radius = body_Gmass.copy()
for k,v in body_Gmass.items():
    body_radius[k] = [((Gmass/GU)/(4./3.*np.pi*density))**(1./3.) for Gmass in v]

body_radius["hitandrun_disrupt"] = [7e-6, 3.25e-6] 
body_radius["hitandrun_pure"] = [7e-6, 3.25e-6] 

# ----------------------------------------------------------------------------------------------------------------------
# Define the animation class that will generate the movies of the fragmentation outcomes
# ----------------------------------------------------------------------------------------------------------------------


def encounter_combiner(sim):
    """
    Combines simulation data with encounter data to produce a dataset that contains the position,
    mass, radius, etc. of both. It will interpolate over empty time values to fill in gaps.
    """

    # Only keep a minimal subset of necessary data from the simulation and encounter datasets
    keep_vars = ['name','rh','vh','Gmass','radius','rot']
    data = sim.data[keep_vars]
    enc = sim.encounters[keep_vars].load()

    # Remove any encounter data at the same time steps that appear in the data to prevent duplicates
    t_not_duplicate = ~data['time'].isin(enc['time'])
    data = data.sel(time=t_not_duplicate)
    tgood=enc.time.where(~np.isnan(enc.time),drop=True)
    enc = enc.sel(time=tgood)

    # The following will combine the two datasets along the time dimension, sort the time dimension, and then fill in any time gaps with interpolation
    ds = xr.combine_nested([data,enc],concat_dim='time').sortby("time").interpolate_na(dim="time")
    
    # Rename the merged Target body so that their data can be combined
    tname=[n for n in ds['name'].data if names[0] in n]
    nottname=[n for n in ds['name'].data if names[0] not in n]
    dslist = []
    for n in tname:
        dsnew = ds.sel(name=n)
        dsnew['name'] = names[0]
        dslist.append(dsnew)

    newds = xr.merge(dslist,compat="no_conflicts") 
    ds = xr.combine_nested([ds.sel(name=nottname),newds],concat_dim="name")
   
    # Interpolate in time to make a smooth, constant time step dataset 
    # Add a bit of padding to the time, otherwise there are some issues with the interpolation in the last few frames.
    smooth_time = np.linspace(start=ds.time[0], stop=ds.time[-1], num=int(1.2*num_movie_frames))
    ds = ds.interp(time=smooth_time)
    ds['rotangle'] = xr.zeros_like(ds['rot'])
    ds['rot'] = ds['rot'].fillna(0.0)

    return ds

def center(Gmass, x, y):
    x = x[~np.isnan(x)]
    y = y[~np.isnan(y)]
    Gmass = Gmass[~np.isnan(Gmass)]
    x_com = np.sum(Gmass * x) / np.sum(Gmass)
    y_com = np.sum(Gmass * y) / np.sum(Gmass)
    return x_com, y_com 
class AnimatedScatter(object):
    """An animated scatter plot using matplotlib.animations.FuncAnimation."""

    def __init__(self, sim, animfile, title, style, nskip=1):
        self.sim = sim
        self.ds = encounter_combiner(sim)
        self.npl = len(self.ds['name']) - 1
        self.title = title
        self.body_color_list = {'Initial conditions': 'xkcd:windows blue',
                      'Disruption': 'xkcd:baby poop',
                      'Supercatastrophic': 'xkcd:shocking pink',
                      'Hit and run fragmention': 'xkcd:blue with a hint of purple',
                      'Central body': 'xkcd:almost black'}

        # Set up the figure and axes...
        self.figsize = (4,4)
        self.fig, self.ax = self.setup_plot()
        self.ani = animation.FuncAnimation(self.fig, self.update_plot, init_func=self.init_func, interval=1, frames=range(0,num_movie_frames,nskip), blit=True)
        self.ani.save(animfile, fps=60, dpi=300, extra_args=['-vcodec', 'libx264'])
        print(f"Finished writing {animfile}")

    def setup_plot(self):
        fig = plt.figure(figsize=self.figsize, dpi=300)
        plt.tight_layout(pad=0)

        # Calculate the distance along the y-axis between the colliding bodies at the start of the simulation.
        # This will be used to scale the axis limits on the movie.
        rhy1 = self.sim.data['rh'].sel(name=names[0],space='y').isel(time=0).values[()]
        rhy2 = self.sim.data['rh'].sel(name=names[1],space='y').isel(time=0).values[()]

        scale_frame =   abs(rhy1) + abs(rhy2)
        if "hitandrun" in style:
           scale_frame *= 2
           
        ax = plt.Axes(fig, [0.1, 0.1, 0.8, 0.8])
        self.ax_pt_size = self.figsize[0] *  72 / scale_frame  * 0.7
        ax.set_xlim(-scale_frame, scale_frame)
        ax.set_ylim(-scale_frame, scale_frame)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_title(self.title)
        fig.add_axes(ax)

        return fig, ax
    
    def init_func(self):
        self.artists = [] 
        aarg = self.vec_props('xkcd:beige')
        for i in range(self.npl):
            self.artists.append(self.ax.annotate("",xy=(0,0),**aarg)) 
        
        self.artists.append(self.ax.scatter([],[],c='k', animated=True, zorder=10))
        return self.artists

    def update_plot(self, frame):
        # Define a function to calculate a reference frame for the animation
        # This will be based on the initial velocity of the Target body

        t, Gmass, rh, radius, rotangle = next(self.data_stream(frame))
        x_ref, y_ref = center(Gmass, rh[:,0], rh[:,1]) 
        rh = np.c_[rh[:,0] - x_ref, rh[:,1] - y_ref]
        self.artists[-1].set_offsets(rh)
        point_rad = radius * self.ax_pt_size
        self.artists[-1].set_sizes(point_rad**2)
        
        sarrowend, sarrowtip = self.spin_arrows(rh, rotangle, 1.1*radius) 
        for i, s in enumerate(self.artists[:-1]):
            self.artists[i].set_position(sarrowtip[i])
            self.artists[i].xy = sarrowend[i]
        
        return self.artists

    def data_stream(self, frame=0):
        while True:
            t = self.ds.isel(time=frame)['time'].values[()]
            
            if frame > 0:
                dsold = self.ds.isel(time=frame-1)
                told = dsold['time'].values[()]
                dt = t - told
                self.ds['rotangle'][dict(time=frame)] = dsold['rotangle'] + dt * dsold['rot']
            
            ds = self.ds.isel(time=frame)
            ds = ds.where(ds['name'] != "Sun", drop=True)
            radius = ds['radius'].values
            Gmass = ds['Gmass'].values
            rh = ds['rh'].values
            rotangle = ds['rotangle'].values

            yield t, Gmass, rh, radius, rotangle
            
    def spin_arrows(self, rh, rotangle, rotlen):
        px = rh[:, 0]
        py = rh[:, 1]
        sarrowend = []
        sarrowtip = []
        for i in range(rh.shape[0]):
            endrel = np.array([0.0, -rotlen[i],  0.0])
            tiprel = np.array([0.0, rotlen[i], 0.0])
            r = R.from_rotvec(rotangle[i,:], degrees=True)
            endrel = r.apply(endrel)
            tiprel = r.apply(tiprel)
            send = (px[i] + endrel[0], py[i] + endrel[1])
            stip = (px[i] + tiprel[0], py[i] + tiprel[1])
            sarrowend.append(send)
            sarrowtip.append(stip)
        return sarrowend, sarrowtip
    
    def vec_props(self, c):
        arrowprops = {
            'arrowstyle': '-',
            'linewidth' : 1,
        }

        arrow_args = {
            'xycoords': 'data',
            'textcoords': 'data',
            'arrowprops': arrowprops,
            'annotation_clip': True,
            'zorder': 100,
            'animated' : True
        }
        aarg = arrow_args.copy()
        aprop = arrowprops.copy()
        aprop['color'] = c
        aarg['arrowprops'] = aprop
        aarg['color'] = c
        return aarg    
    
if __name__ == "__main__":

    print("Select a fragmentation movie to generate.")
    print("1. Head-on disruption")
    print("2. Off-axis disruption")
    print("3. Head-on supercatastrophic")
    print("4. Off-axis supercatastrophic")
    print("5. Hit and run with disruption of the runner")
    print("6. Pure hit and run")
    print("7. Merge")
    print("8. Merge crossing the spin barrier")
    print("9. All of the above")
    user_selection = int(input("? "))

    if user_selection > 0 and user_selection < 9:
        movie_styles = [available_movie_styles[user_selection-1]]
    else:
        print("Generating all movie styles")
        movie_styles = available_movie_styles.copy()

    for style in movie_styles:
        print(f"Generating {movie_titles[style]}")
        movie_filename = f"{style}.mp4"
        # Pull in the Swiftest output data from the parameter file and store it as a Xarray dataset.
        sim = swiftest.Simulation(simdir=style, rotation=True, init_cond_format = "XV", compute_conservation_values=True)
        sim.add_solar_system_body("Sun")
        sim.add_body(name=names, Gmass=body_Gmass[style], radius=body_radius[style], rh=pos_vectors[style], vh=vel_vectors[style], rot=rot_vectors[style])

        # Set fragmentation parameters
        minimum_fragment_gmass = 0.01 * body_Gmass[style][1] 
        gmtiny = 0.50 * body_Gmass[style][1] 
        sim.set_parameter(collision_model="fraggle", 
                          encounter_save="both", 
                          gmtiny=gmtiny, 
                          minimum_fragment_gmass=minimum_fragment_gmass, 
                          nfrag_reduction=nfrag_reduction[style])
        sim.run(dt=5e-4, tstop=tstop[style], istep_out=1, dump_cadence=0)

        print("Generating animation")
        anim = AnimatedScatter(sim,movie_filename,movie_titles[style],style,nskip=1)



#!/usr/bin/env python3

"""
 Copyright 2024 - The Minton Group at Purdue University
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

"""
Generates and runs a set of Swiftest input files from initial conditions for the Spherical Harmonics features with the 
SyMBA integrator. Using Chariklo as the example body with axes measurements taken from Leiva, et al (2017) (Jacobi 
Ellipsoid model). All simulation outputs are stored in the /simdata subdirectory.

"""

import swiftest
import numpy as np

seed = 123
rng = np.random.default_rng(seed=seed)

# set up swiftest simulation with relevant units (here they are km, days, and kg)
sim = swiftest.Simulation(DU2M = 1e3, TU = 'd', MU = 'kg')
sim.clean()

# Central Body Parameters (Chariklo parameters from Leiva, et al (2017) (Jacobi Ellipsoid model))
cb_mass = 6.1e18 # kg
cb_radius = 123 # km
cb_a = 157 # km
cb_b = 139 # km
cb_c = 86 # km
cb_volume = 4.0 / 3 * np.pi * cb_radius**3 # km^3
cb_density = cb_mass / cb_volume 
cb_T_rotation = 7.004 / 24.0 # converting from hours to julian days (TU)
cb_rot = [[0, 0, 360.0 / cb_T_rotation]] # degrees/d

# Extract the spherical harmonics coefficients (c_lm) from axes measurements
#
# The user can pass an optional reference radius at which the coefficients are calculated. If not provided, SHTOOLS 
# calculates the reference radius. If lref_radius = True, the function returns the reference radius used. 
# We recommend setting passing and setting a reference radius. Coefficients are geodesy (4-pi) normalised.

c_lm, cb_radius = swiftest.clm_from_ellipsoid(mass = cb_mass, density = cb_density, a = cb_a, b = cb_b, c = cb_c, lmax = 6, lref_radius = True, ref_radius = cb_radius)

# Add the central body
# The user can pass the c_lm coefficients directly to the add_body method if they do not wish to use the clm_from_ellipsoid method.
sim.add_body(name = 'Chariklo', mass = cb_mass, rot = cb_rot, radius = cb_radius, c_lm = c_lm)

# Add user-defined massive bodies
npl         = 5
density_pl  = cb_density

name_pl     = ["SemiBody_01", "SemiBody_02", "SemiBody_03", "SemiBody_04", "SemiBody_05"]
a_pl        = rng.uniform(250, 400, npl)
e_pl        = rng.uniform(0.0, 0.05, npl)
inc_pl      = rng.uniform(0.0, 10, npl)
capom_pl    = rng.uniform(0.0, 360.0, npl)
omega_pl    = rng.uniform(0.0, 360.0, npl)
capm_pl     = rng.uniform(0.0, 360.0, npl)
R_pl        = np.array([0.5, 1.0, 1.2, 0.75, 0.8])
M_pl        = 4.0 / 3 * np.pi * R_pl**3 * density_pl
Ip_pl       = np.full((npl,3),0.4,)
rot_pl      = np.zeros((npl,3))
mtiny       = 1.1 * np.max(M_pl)

sim.add_body(name=name_pl, a=a_pl, e=e_pl, inc=inc_pl, capom=capom_pl, omega=omega_pl, capm=capm_pl, mass=M_pl, radius=R_pl,  Ip=Ip_pl, rot=rot_pl)

# Add 10 user-defined test particles.
ntp = 10

name_tp     = ["TestParticle_01", "TestParticle_02", "TestParticle_03", "TestParticle_04", "TestParticle_05", "TestParticle_06", "TestParticle_07", "TestParticle_08", "TestParticle_09", "TestParticle_10"]
a_tp        = rng.uniform(250, 400, ntp)
e_tp        = rng.uniform(0.0, 0.05, ntp)
inc_tp      = rng.uniform(0.0, 10, ntp)
capom_tp    = rng.uniform(0.0, 360.0, ntp)
omega_tp    = rng.uniform(0.0, 360.0, ntp)
capm_tp     = rng.uniform(0.0, 360.0, ntp)

sim.add_body(name=name_tp, a=a_tp, e=e_tp, inc=inc_tp, capom=capom_tp, omega=omega_tp, capm=capm_tp)
sim.set_parameter(tstart=0.0, tstop=10.0, dt=0.01, istep_out=10, dump_cadence=0, compute_conservation_values=True, mtiny=mtiny)

# Display the run configuration parameters.
sim.get_parameter()

# Run the simulation. Arguments may be defined here or thorugh the swiftest.Simulation() method.
sim.run()



#!/usr/bin/env python3
"""
 Copyright 2023 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

"""
Generates, runs, and processes a set of initial conditions for a multi-body super-catastrophic distruption collisional event. 
All Swiftest output files are stored in the /supercatastrophic_multi subdirectory.

**NOTE: You must have ffmpeg installed on your system before running this script. For instance, on MacOS:

```brew install ffmpeg```

on Ubuntu:

```sudo apt-get install ffmpeg```


Inputs
_______
None.

Returns
-------
supercatastrophic_multi.mp4 : A .mp4 file of a fragmentation event.
collisions.log   : An ASCII file containing the information of any collisional events that occured.
collisions.nc    : A NetCDF file containing the collision output.
data.nc          : A NetCDF file containing the simulation output.
encounters.nc    : A NetCDF file containing the encounters output.
init_cond.nc     : A NetCDF file containing the initial conditions for the simulation.
param.00....in   : A series of parameter input files containing the parameters for the simulation at every output stage.
param.in         : An ASCII file containing the inital parameters for the simulation.
param.restart.in : An ASCII file containing the parameters for the simulation at the last output. 
swiftest.log     : An ASCII file containing the information on the status of the simulation as it runs.
"""

import swiftest
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from scipy.spatial.transform import Rotation as R

# ----------------------------------------------------------------------------------------------------------------------
# Define the names and initial conditions of the various fragmentation simulation types
# ----------------------------------------------------------------------------------------------------------------------
available_movie_styles = ["supercatastrophic_multi"]
movie_title_list = ["Multi-body Supercatastrophic"]
movie_titles = dict(zip(available_movie_styles, movie_title_list))
num_movie_frames = 1000


# ----------------------------------------------------------------------------------------------------------------------
# To increase the number of bodies generated in each collision type, decrease the value of the corresponding nfrag_reduction number 
# ----------------------------------------------------------------------------------------------------------------------
nfrag_reduction = {
    "supercatastrophic_multi" : 5.0,
    }



# These initial conditions were generated by trial and error
names = ["Body1","Body2","Body3","Body4"]

rdistance = 1e-4/np.sqrt(2.0)
vrel = 2.0/np.sqrt(2.0)

rcb = np.array([1.0, 0.0,  0.0])
vcb = np.array([0.0, 6.28, 0.0])

r1 = rcb + np.array([-rdistance,-rdistance,0.0])
r2 = rcb + np.array([ rdistance,-rdistance,0.0])
r3 = rcb + np.array([ rdistance, rdistance,0.0])
r4 = rcb + np.array([-rdistance, rdistance,0.0])

v1 = vcb + np.array([ vrel, vrel, 0.0])
v2 = vcb + np.array([-vrel, vrel, 0.0])
v3 = vcb + np.array([-vrel,-vrel, 0.0])
v4 = vcb + np.array([ vrel,-vrel, 0.0])

rot1 = np.array([0.0,0.0,1e5])
rot2 = np.array([0.0,0.0,-2e5])
rot3 = np.array([0.0,0.0,3e5])
rot4 = np.array([0.0,0.0,0.0])

m1 = 2e-7
m2 = 2e-7
m3 = 2e-7
m4 = 2e-7

pos_vectors = {"supercatastrophic_multi"   : [r1, r2, r3, r4],
               }

vel_vectors = {"supercatastrophic_multi":   [v1, v2, v3, v4]
               }

rot_vectors = { "supercatastrophic_multi":   [rot1, rot2, rot3, rot4]
               }

body_Gmass = {"supercatastrophic_multi":   [m1, m2, m3, m4]
               }

tstop = {"supercatastrophic_multi"   : 2.0e-3,
         }

density = 3000 * swiftest.AU2M**3 / swiftest.MSun
GU = swiftest.GMSun * swiftest.YR2S**2 / swiftest.AU2M**3
body_radius = body_Gmass.copy()
for k,v in body_Gmass.items():
    body_radius[k] = [((Gmass/GU)/(4./3.*np.pi*density))**(1./3.) for Gmass in v]

# ----------------------------------------------------------------------------------------------------------------------
# Define the animation class that will generate the movies of the fragmentation outcomes
# ----------------------------------------------------------------------------------------------------------------------
def encounter_combiner(sim):
    """
    Combines simulation data with encounter data to produce a dataset that contains the position,
    mass, radius, etc. of both. It will interpolate over empty time values to fill in gaps.
    """

    # Only keep a minimal subset of necessary data from the simulation and encounter datasets
    keep_vars = ['name','rh','vh','Gmass','radius','rot']
    data = sim.data[keep_vars]
    enc = sim.encounters[keep_vars].load()

    # Remove any encounter data at the same time steps that appear in the data to prevent duplicates
    t_not_duplicate = ~enc['time'].isin(data['time'])
    enc = enc.where(t_not_duplicate,drop=True)
    tgood=enc.time.where(~np.isnan(enc.time),drop=True)
    enc = enc.sel(time=tgood)

    # The following will combine the two datasets along the time dimension, sort the time dimension, and then fill in any time gaps with interpolation
    ds = xr.combine_nested([data,enc],concat_dim='time').sortby("time").interpolate_na(dim="time")
    
    # Interpolate in time to make a smooth, constant time step dataset 
    # Add a bit of padding to the time, otherwise there are some issues with the interpolation in the last few frames.
    smooth_time = np.linspace(start=tgood.isel(time=0), stop=ds.time[-1], num=int(1.2*num_movie_frames))
    ds = ds.interp(time=smooth_time)
    ds['rotangle'] = xr.zeros_like(ds['rot'])
    ds['rot'] = ds['rot'].fillna(0.0)

    return ds

def center(Gmass, x, y):
    x = x[~np.isnan(x)]
    y = y[~np.isnan(y)]
    Gmass = Gmass[~np.isnan(Gmass)]
    x_com = np.sum(Gmass * x) / np.sum(Gmass)
    y_com = np.sum(Gmass * y) / np.sum(Gmass)
    return x_com, y_com 
class AnimatedScatter(object):
    """An animated scatter plot using matplotlib.animations.FuncAnimation."""

    def __init__(self, sim, animfile, title, style, nskip=1):

        self.ds = encounter_combiner(sim)
        self.npl = len(self.ds['name']) - 1
        self.title = title
        self.body_color_list = {'Initial conditions': 'xkcd:windows blue',
                      'Disruption': 'xkcd:baby poop',
                      'Supercatastrophic': 'xkcd:shocking pink',
                      'Hit and run fragmention': 'xkcd:blue with a hint of purple',
                      'Central body': 'xkcd:almost black'}

        # Set up the figure and axes...
        self.figsize = (4,4)
        self.fig, self.ax = self.setup_plot()
        self.ani = animation.FuncAnimation(self.fig, self.update_plot, init_func=self.init_func, interval=1, frames=range(0,num_movie_frames,nskip), blit=True)
        self.ani.save(animfile, fps=60, dpi=300, extra_args=['-vcodec', 'libx264'])
        print(f"Finished writing {animfile}")

    def setup_plot(self):
        fig = plt.figure(figsize=self.figsize, dpi=300)
        plt.tight_layout(pad=0)

        # Calculate the distance along the y-axis between the colliding bodies at the start of the simulation.
        # This will be used to scale the axis limits on the movie.
        rhy1 = self.ds['rh'].sel(name="Body1",space='y').isel(time=0).values[()]
        rhy2 = self.ds['rh'].sel(name="Body3",space='y').isel(time=0).values[()]

        scale_frame =   abs(rhy1) + abs(rhy2)
           
        ax = plt.Axes(fig, [0.1, 0.1, 0.8, 0.8])
        self.ax_pt_size = self.figsize[0] *  72 / scale_frame  * 0.7
        ax.set_xlim(-scale_frame, scale_frame)
        ax.set_ylim(-scale_frame, scale_frame)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_title(self.title)
        fig.add_axes(ax)

        return fig, ax
    
    def init_func(self):
        self.artists = [] 
        aarg = self.vec_props('xkcd:beige')
        for i in range(self.npl):
            self.artists.append(self.ax.annotate("",xy=(0,0),**aarg)) 
        
        self.artists.append(self.ax.scatter([],[],c='k', animated=True, zorder=10))
        return self.artists

    def update_plot(self, frame):
        # Define a function to calculate a reference frame for the animation
        t, Gmass, rh, radius, rotangle = next(self.data_stream(frame))
        x_ref, y_ref = center(Gmass, rh[:,0], rh[:,1]) 
        rh = np.c_[rh[:,0] - x_ref, rh[:,1] - y_ref]
        self.artists[-1].set_offsets(rh)
        point_rad = radius * self.ax_pt_size
        self.artists[-1].set_sizes(point_rad**2)
        
        sarrowend, sarrowtip = self.spin_arrows(rh, rotangle, 1.1*radius) 
        for i, s in enumerate(self.artists[:-1]):
            self.artists[i].set_position(sarrowtip[i])
            self.artists[i].xy = sarrowend[i]
        
        return self.artists

    def data_stream(self, frame=0):
        while True:
            t = self.ds.isel(time=frame)['time'].values[()]
            
            if frame > 0:
                dsold = self.ds.isel(time=frame-1)
                told = dsold['time'].values[()]
                dt = t - told
                self.ds['rotangle'][dict(time=frame)] = dsold['rotangle'] + dt * dsold['rot']
            
            ds = self.ds.isel(time=frame)
            ds = ds.where(ds['name'] != "Sun", drop=True)
            radius = ds['radius'].values
            Gmass = ds['Gmass'].values
            rh = ds['rh'].values
            rotangle = ds['rotangle'].values

            yield t, Gmass, rh, radius, rotangle
            
    def spin_arrows(self, rh, rotangle, rotlen):
        px = rh[:, 0]
        py = rh[:, 1]
        sarrowend = []
        sarrowtip = []
        for i in range(rh.shape[0]):
            endrel = np.array([0.0, -rotlen[i],  0.0])
            tiprel = np.array([0.0, rotlen[i], 0.0])
            r = R.from_rotvec(rotangle[i,:], degrees=True)
            endrel = r.apply(endrel)
            tiprel = r.apply(tiprel)
            send = (px[i] + endrel[0], py[i] + endrel[1])
            stip = (px[i] + tiprel[0], py[i] + tiprel[1])
            sarrowend.append(send)
            sarrowtip.append(stip)
        return sarrowend, sarrowtip
    
    def vec_props(self, c):
        arrowprops = {
            'arrowstyle': '-',
            'linewidth' : 1,
        }

        arrow_args = {
            'xycoords': 'data',
            'textcoords': 'data',
            'arrowprops': arrowprops,
            'annotation_clip': True,
            'zorder': 100,
            'animated' : True
        }
        aarg = arrow_args.copy()
        aprop = arrowprops.copy()
        aprop['color'] = c
        aarg['arrowprops'] = aprop
        aarg['color'] = c
        return aarg    
    
if __name__ == "__main__":

    print("Select a fragmentation movie to generate.")
    print("1. Multi-body supercatastrophic")
    print("2. All of the above")
    user_selection = int(input("? "))

    if user_selection > 0 and user_selection < 2:
        movie_styles = [available_movie_styles[user_selection-1]]
    else:
        print("Generating all movie styles")
        movie_styles = available_movie_styles.copy()

    for style in movie_styles:
        print(f"Generating {movie_titles[style]}")
        movie_filename = f"{style}.mp4"
        # Pull in the Swiftest output data from the parameter file and store it as a Xarray dataset.
        sim = swiftest.Simulation(simdir=style, rotation=True, init_cond_format = "XV", compute_conservation_values=True)
        sim.add_solar_system_body("Sun")
        sim.add_body(name=names, Gmass=body_Gmass[style], radius=body_radius[style], rh=pos_vectors[style], vh=vel_vectors[style], rot=rot_vectors[style])

        # Set fragmentation parameters
        minimum_fragment_gmass = 0.05 * body_Gmass[style][1] 
        gmtiny = 0.10 * body_Gmass[style][1] 
        sim.set_parameter(collision_model="fraggle", 
                          encounter_save="both", 
                          gmtiny=gmtiny, 
                          minimum_fragment_gmass=minimum_fragment_gmass,
                          nfrag_reduction=nfrag_reduction[style],
                          verbose=False)
        sim.run(dt=5e-4, tstop=tstop[style], istep_out=1, dump_cadence=0)

        print("Generating animation")
        anim = AnimatedScatter(sim,movie_filename,movie_titles[style],style,nskip=1)



#!/usr/bin/env python3

"""
 Copyright 2023 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

"""
Creates a movie from a set of Swiftest output files. All simulation 
outputs are stored in the /simdata subdirectory.

**NOTE: You must have ffmpeg installed on your system before running this script. For instance, on MacOS:

```brew install ffmpeg```

on Ubuntu:

```sudo apt-get install ffmpeg```


Input
------
param.in    : ASCII Swiftest parameter input file.
data.nc     : A NetCDF file containing the simulation output.

Output
------
Chambers2013-aescatter.mp4  : A .mp4 file plotting eccentricity vs semimajor axis. 
OR 
Chambers2013-aiscatter.mp4  : A .mp4 file plotting inclination vs semimajor axis. 
"""

import swiftest 
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
import matplotlib.colors as mcolors
from collections import namedtuple


plt.switch_backend('agg')

titletext = "Chambers (2013)"
valid_plot_styles = ["aescatter", "aiscatter", "arotscatter"]
xlim={"aescatter" : (0.0, 3.0),
      "aiscatter" : (0.0, 3.0), 
      "arotscatter" : (0.0, 2.5)}
ylim={"aescatter" : (0.0, 1.0),
      "aiscatter" : (0.0, 40.0),
      "arotscatter" : (1.0, 10000.0)}
xlabel={"aescatter": "Semimajor axis (AU)",
        "aiscatter": "Semimajor axis (AU)",
        "arotscatter": "Semimajor axis (AU)"}
ylabel={"aescatter": "Eccentricity",
        "aiscatter": "Inclination (deg)",
        "arotscatter": "Rotation period (h)"}

YR2HR = 365.25 * 24
ROT2PERIOD = YR2HR * 360.0

framejump = 1
origin_types = ["Initial conditions", "Merger", "Disruption", "Supercatastrophic", "Hit and run fragmentation"]


class AnimatedScatter(object):
    """An animated scatter plot using matplotlib.animations.FuncAnimation."""
    def __init__(self, ds, param):

        self.radscale = 2000
        nframes = int(ds['time'].size / framejump)
        ds['rot_mag'] = swiftest.tool.magnitude(ds,"rot")
        ds['rot_mag'] = ROT2PERIOD / ds['rot_mag']
        self.Rcb = ds['radius'].sel(name="Sun").isel(time=0).values[()]
        self.ds = ds
        self.param = param
        colors = ["k", "xkcd:faded blue", "xkcd:marigold", "xkcd:shocking pink", "xkcd:baby poop green"]
        self.clist = dict(zip(origin_types,colors))

        # Setup the figure and axes...
        fig = plt.figure(figsize=(8,4.5), dpi=300)
        plt.tight_layout(pad=0)
        # set up the figure
        self.ax = plt.Axes(fig, [0.1, 0.15, 0.8, 0.75])
        fig.add_axes(self.ax)

        self.make_artists()
        
        self.ani = animation.FuncAnimation(fig, func=self.update, interval=1, frames=nframes, init_func=self.init_plot, blit=True)
        self.ani.save(animation_file, fps=60, dpi=300, extra_args=['-vcodec', 'libx264'])
        print(f'Finished writing {animation_file}')
        
    def make_artists(self):
        scatter_names = [f"s{i}" for i,k in enumerate(origin_types)]
        self.scatter_artist_names = dict(zip(origin_types,scatter_names))
        
        animated_elements = [self.ax.text(0.50, 1.05, "", bbox={'facecolor': 'w', 'alpha': 0.5, 'pad': 5}, transform=self.ax.transAxes,ha="center", animated=True)]
        element_names = ["title"]
        for key, value in self.clist.items():
            animated_elements.append(self.ax.scatter([], [], marker='o', s=[], c=value, alpha=0.75, label=key, animated=True))
            element_names.append(self.scatter_artist_names[key])
        
        Artists = namedtuple("Artists",tuple(element_names))
        self.artists = Artists(*animated_elements)
        return 

    def init_plot(self):
        self.ax.set_xlim(xlim[plot_style])
        self.ax.set_ylim(ylim[plot_style])
        
        # set up the figure
        self.ax.margins(x=10, y=1)
        self.ax.set_xlabel(xlabel[plot_style], fontsize='16', labelpad=1)
        self.ax.set_ylabel(ylabel[plot_style], fontsize='16', labelpad=1) 
        
        leg = plt.legend(loc="upper left", scatterpoints=1, fontsize=10)
        for i,l in enumerate(leg.legend_handles):
            leg.legend_handles[i]._sizes = [20]
        
        if plot_style == "arotscatter":
            self.ax.set_yscale('log')
        
        return self.artists

    def get_data(self, frame=0):
        d = self.ds.isel(time = frame)
        n=len(d['name'])
        d = d.isel(name=range(1,n))
        d['radmarker'] = (d['radius'] / self.Rcb) * self.radscale

        t = d['time'].values
        npl = d['npl'].values
        radmarker = d['radmarker'].values
        origin = d['origin_type'].values
        
        if plot_style == "aescatter":
            pl = np.c_[d['a'].values,d['e'].values]
        elif plot_style == "aiscatter":
            pl = np.c_[d['a'].values,d['inc'].values]
        elif plot_style == "arotscatter":
            pl = np.c_[d['a'].values,d['rot_mag'].values]

        return t, npl, pl, radmarker, origin

    def update(self,frame):
        """Update the scatter plot."""
        t,  npl, pl, radmarker, origin = self.get_data(framejump * frame)

        self.artists.title.set_text(f"{titletext} - Time = ${t*1e-6:6.3f}$ My with ${npl:4.0f}$ particles")

        for key,name in self.scatter_artist_names.items():
            idx = origin == key
            if any(idx) and any(~np.isnan(radmarker[idx])):
                scatter = self.artists._asdict()[name]
                scatter.set_sizes(radmarker[idx])
                scatter.set_offsets(pl[idx,:])
                scatter.set_facecolor(self.clist[key])

        return self.artists

sim = swiftest.Simulation(read_data=True, read_collisions=False, dask=True)
for plot_style in valid_plot_styles:
    animation_file = f"Chambers2013-{plot_style}.mp4"
    print('Making animation')
    anim = AnimatedScatter(sim.data,sim.param)
    print('Animation finished')



#!/usr/bin/env python3

"""
 Copyright 2023 - David Minton
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

"""
This will run the simulation from a set of initial conditions. The simulation parameters in this file are set to generate
a very short simulation for testing purposes. Edit the values passed to the run() function as necessary.

Input
------
simdata/param.in    : ASCII Swiftest parameter input file.

Output
------
Outputs are stored in the /simdata subdirectory.

"""
import swiftest
sim = swiftest.Simulation(read_param=True)

# Original run parameters
# sim.run(tstop=3e8, dt=6.0875/365.25, istep_out=60000, dump_cadence=10,integreator="symba")
# 
sim.run(tstop=10000.0, dt=6.0875/365.25, istep_out=10000, dump_cadence=0, integrator="symba")



#!/usr/bin/env python3

"""
 Copyright 2023 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

"""
Generates a set of Swiftest input files from initial conditions with the SyMBA integrator. All simulation 
outputs are stored in the /simdata subdirectory.

Input
------
None.

Output
------
inital_conditions.png : A .png file depicting the simulation initial configuration.
init_cond.nc          : A NetCDF file containing the initial conditions for the simulation.
param.in              : An ASCII file containing the parameters for the simulation.
""" 

import swiftest
import numpy as np
from numpy.random import default_rng
import matplotlib.pyplot as plt

# Initialize simulation object
sim = swiftest.Simulation(compute_conservation_values=True, rotation=True, init_cond_format="EL",collision_model="fraggle",encounter_save="none")
sim.clean()

# Add bodies described in Chambers (2013) Sec. 2.1, with the uniform spatial distribution and two bodies sizes (big and small)
Nb = 14
Ns = 140
Mb = 2.8e-7 * 14 / Nb
Ms = 2.8e-8 * 140 / Ns
dens = 3000.0 * sim.KG2MU / sim.M2DU**3
rot = 1e-6 / sim.TU2S # Use a small but non-zero value for the initial rotation state to prevent divide-by-zero errors in analysis

mtiny = 1e-2 * Ms
minimum_fragment_mass = 1e-5 * Ms
nfrag_reduction = 30.0
rng = default_rng(seed=3031179)

runname = "Chambers (2013)"
def a_profile(n_bodies):
    """
    Generates the profile described in Sec. 2.1 of Chambers:  
    
    *In all cases, the surface density R = 8 g/cm2 at 1 AU, varying as a**(-3/2), where a is the orbital semi-major axis. 
    The region with a < 0.7 AU deviates from this law, declining linearly with decreasing distance until R = 0 at 0.3 AU. 
    The outer edge of the disk is 2 AU in all cases.
    """
    def sample(r_inner, r_break, r_outer, slope1, slope2, size):
        """ 
        Define the functions to pull random semi-major axes from a distribution using a rejection sampling method
        This defines a 2-slope model with a break at r_break
        Based on (https://stackoverflow.com/questions/66874819/random-numbers-with-user-defined-continuous-probability-distribution)
        """
        y=np.ones([size])
        pdf=np.zeros([size])
        x=np.empty_like(y)
        while np.any(y>pdf): 
            x = np.where(y>pdf,rng.uniform(r_inner, r_outer, size=size),x)
          
            # The proportionality factor A ensures that the PDF approaches the same value on either side of the break point
            # Assumes the break point is the max of the PDF
            A=np.where(x < r_break,1.0,r_break**(slope1-slope2))
            slope=np.where(x < r_break,slope1+1,slope2+1)
            y = np.where(y>pdf,rng.uniform(0, A*r_break**slope, size=size),y)
            pdf = np.where(y>pdf,A*x**(slope),pdf)
        return x
    
    a_inner = 0.3
    a_break = 0.7
    a_outer = 2.0
    slope1 = 1.0
    slope2 = -1.5
   
    a_vals = sample(a_inner, a_break, a_outer, slope1, slope2, n_bodies)
    return a_vals

# Define the initial orbital elements of the big and small bodies
avalb = a_profile(Nb)
avals = a_profile(Ns)

esigma = 0.01
isigma = np.rad2deg(0.5 * esigma)
evalb = rng.rayleigh(scale=esigma, size=Nb)
evals = rng.rayleigh(scale=esigma, size=Ns)
incvalb = rng.rayleigh(scale=isigma, size=Nb)
incvals = rng.rayleigh(scale=isigma, size=Ns)

capomvalb = rng.uniform(0.0, 360.0, Nb)
capomvals = rng.uniform(0.0, 360.0, Ns)
omegavalb = rng.uniform(0.0, 360.0, Nb)
omegavals = rng.uniform(0.0, 360.0, Ns)
capmvalb = rng.uniform(0.0, 360.0, Nb)
capmvals = rng.uniform(0.0, 360.0, Ns)
Ipvalb = np.full((Nb,3), 0.4)
Ipvals = np.full((Ns,3), 0.4)
rotvalb = np.full_like(Ipvalb,rot)
rotvals = np.full_like(Ipvals,rot)

noise_digits = 4 # Approximately the number of digits of precision to vary the mass values to avoid duplicate masses
epsilon = np.finfo(float).eps
Mnoiseb = 1.0 + 10**noise_digits * rng.uniform(-epsilon,epsilon, Nb)
Mnoises = 1.0 + 10**noise_digits * rng.uniform(-epsilon,epsilon, Ns)

Mvalb = Mb * Mnoiseb
Mvals = Ms * Mnoises

Rvalb = (3 * Mvalb / (4 * np.pi * dens) )**(1.0 / 3.0)
Rvals = (3 * Mvals / (4 * np.pi * dens) )**(1.0 / 3.0)

# Give the bodies unique names
nameb = [f"Big{i:03}" for i in range(Nb)]
names = [f"Small{i:03}" for i in range(Ns)]

# Add the modern planets and the Sun using the JPL Horizons Database.
sim.add_solar_system_body(["Sun","Jupiter","Saturn","Uranus","Neptune"])
sim.add_body(name=nameb, a=avalb, e=evalb, inc=incvalb, capom=capomvalb, omega=omegavalb, capm=capmvalb, mass=Mvalb, radius=Rvalb, rot=rotvalb, Ip=Ipvalb)
sim.add_body(name=names, a=avals, e=evals, inc=incvals, capom=capomvals, omega=omegavals, capm=capmvals, mass=Mvals, radius=Rvals, rot=rotvals, Ip=Ipvals)
sim.set_parameter(mtiny=mtiny, minimum_fragment_mass=minimum_fragment_mass, nfrag_reduction=nfrag_reduction)

sim.set_parameter(tstop=3e8, dt=6.0875/365.25, istep_out=60000, dump_cadence=10)
sim.save()

# Plot the initial conditions
fig = plt.figure(figsize=(8.5, 11))
ax1 = plt.subplot(2, 1, 1)
ax2 = plt.subplot(2, 1, 2)
fig.suptitle(runname)

ic = sim.init_cond.isel(time=0)
radius = ic['radius'].values
markersize = radius * 4e5
markercolor = np.where(radius < 2.0e-5, np.full_like(markersize, "black",dtype=object), np.full_like(markersize,"red",dtype=object))
a = ic['a'].values
e = ic['e'].values
inc = ic['inc'].values

ax1.scatter(a, e, s=markersize, color=markercolor, label=None)
ax1.set_xlabel("Semimajor Axis (AU)")
ax1.set_ylabel("Eccentricity")
ax1.set_xlim([0.0, 2.5])
ax1.set_ylim([0.0,4*esigma])

ax2.scatter(a, inc, s=markersize, color=markercolor, label=None)
ax2.set_xlabel("Semimajor Axis (AU)")
ax2.set_ylabel("Inclination (deg)")
ax2.set_xlim([0.0, 2.5])
ax2.set_ylim([0.0,8*isigma])

ax1.scatter(-1, -1, label='planetesimal', color='black')
ax1.scatter(-1, -1, label='embryo', color='red') 
ax1.legend(loc='upper right', frameon=False)

plt.tight_layout()
plt.show()
fig.savefig('initial_conditions.png',dpi=300)




import swiftest
import numpy as np

swiftersim = swiftest.Simulation(simdir="swifter_sim", read_data=True, codename="Swifter")
swiftestsim = swiftest.Simulation(simdir="swiftest_sim",read_data=True)
swiftestsim.data = swiftestsim.data.swap_dims({"name" : "id"})
swiftdiff = swiftestsim.data - swiftersim.data
swiftdiff['rh'].plot(x="time",hue="id",col="space")
swiftdiff['vh'].plot(x="time",hue="id",col="space")


"""
Copyright 2022 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
This file is part of Swiftest.
Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Swiftest. 
If not, see: https://www.gnu.org/licenses. 
"""

import numpy as np
import astropy.constants as const
import astropy.units as u
from astropy.coordinates import SkyCoord

# Constants in SI units
GC = const.G.value[()]
AU2M = const.au.value
GMSun = const.GM_sun.value
MSun = const.M_sun.value
RSun = const.R_sun.value
MEarth = const.M_earth.value
REarth = const.R_earth.value
GMEarth = const.GM_earth.value
JD2S = 86400
YR2S = 365.25 * JD2S
einsteinC = 299792458.0
# Solar oblatenes values: From Mecheri et al. (2004), using Corbard (b) 2002 values (Table II)
J2Sun = 2.198e-7
J4Sun = -4.805e-9
rotpoleSun = SkyCoord(ra=286.13 * u.degree, dec=63.87 * u.degree).cartesian
rotSun = (360.0 / 25.05) / JD2S  * rotpoleSun 




#!/usr/bin/env python3
import numpy as np
import swiftest 

# Simple initial conditions of a circular planet with one test particle in a close encounter state
# Simulation start, stop, and output cadence times
t_0	  = 0 # simulation start time
deltaT	= 0.25 * swiftest.JD2S / swiftest.YR2S   # simulation step size
end_sim = 0.15
t_print = deltaT  #output interval to print results

iout = int(np.ceil(t_print / deltaT))

radius = np.double(4.25875607065041e-05)
Gmass = np.double(0.00012002693582795244940133) 
apl = np.longdouble(1.0)
atp = np.longdouble(1.01)
vpl = np.longdouble(2 * np.pi)
vtp = np.longdouble(2 * np.pi / np.sqrt(atp))

p_pl = np.array([apl, 0.0, 0.0], dtype=np.double)
v_pl = np.array([0.0, vpl, 0.0], dtype=np.double)

p_tp = np.array([atp, 0.0, 0.0], dtype=np.double)
v_tp = np.array([0.0, vtp, 0.0], dtype=np.double)

rhill = np.double(apl * 0.0100447248332378922085)

sim = swiftest.Simulation(simdir="swiftest_sim",init_cond_format="XV", general_relativity=False, integrator="RMVS")
sim.clean()
sim.add_solar_system_body(["Sun"])
sim.add_body(name=["Planet"], rh=[p_pl], vh=[v_pl], Gmass=[Gmass], radius=[radius], rhill=[rhill])
sim.add_body(name=["TestParticle"], rh=[p_tp], vh=[v_tp])
sim.set_parameter(tstart=t_0, tstop=end_sim, dt=deltaT, istep_out=iout, dump_cadence=0)
sim.save()

sim.set_parameter(simdir="swifter_sim",codename="Swifter",init_cond_file_type="ASCII",output_format="XV",output_file_type="REAL8")
sim.save()


"""
Copyright 2024 - The Minton Group at Purdue University
This file is part of Swiftest.
Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Swiftest. 
If not, see: https://www.gnu.org/licenses. 
"""

from .constants import *
from .simulation import Simulation
from .shgrav import clm_from_ellipsoid, clm_from_relief